#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

import copy
import re

from container_app_conf.entry.string import StringConfigEntry

# workaround for deepcopy bug in python<=3.6
# see: https://stackoverflow.com/questions/6279305/typeerror-cannot-deepcopy-this-pattern-object/56935186#56935186
copy._deepcopy_dispatch[type(re.compile(''))] = lambda r, _: r


class RegexConfigEntry(StringConfigEntry):
    _example = r"^[a-zA-z0-9]*$"

    def _value_to_type(self, value: any) -> str or None:
        s = super()._value_to_type(value)
        if s is None and self._none_allowed:
            return None
        return re.compile(s)

    def _type_to_value(self, type: any) -> str:
        return type.pattern



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
import logging
import os
from typing import List

from container_app_conf import ConfigEntry

LOGGER = logging.getLogger(__name__)


class DataSource:

    def has(self, entry: ConfigEntry) -> bool:
        """
        Checks whether the data source has a value for the given config entry
        :param entry: the config entry to check
        :return: True if the source contains a value for the given entry, False otherwise
        """
        raise NotImplementedError()

    def get(self, entry: ConfigEntry) -> any:
        """
        Retrieve the value of the given key
        :param entry: config entry
        :return: value
        """
        raise NotImplementedError()

    def write_reference(self, reference: dict):
        """
        Writes a reference configuration (if implemented)
        :param reference: the reference config as in tree form
        """
        pass


class FilesystemSource(DataSource):
    DEFAULT_FILE_EXTENSIONS = []

    def __init__(self, file_name: str or List[str],
                 path: str or List[str] = None,
                 file_extension: str or List[str] = None):
        """
        :param path: allowed config file path(s)
        :param file_name: allowed config file name(s)
        :param file_extension: allowed config file extension(s)
        """
        self.root = {}
        if path is None:
            from container_app_conf import DEFAULT_CONFIG_FILE_PATHS
            self.paths = DEFAULT_CONFIG_FILE_PATHS
        else:
            self.paths = path if isinstance(path, list) else [path]
        self.file_names = file_name if isinstance(file_name, list) else [file_name]
        if file_extension is None:
            self.file_extensions = self.DEFAULT_FILE_EXTENSIONS
        else:
            self.file_extensions = file_extension if isinstance(file_extension, list) else [file_extension]

    def load(self):
        """
        Loads the config file content into memory
        """
        file_path = self._find_config_file()
        if file_path is None:
            LOGGER.debug("No config file found in paths: {}".format(self.paths))
            return

        self.root = self._load(file_path)

    def _load(self, file_path: str) -> dict:
        """
        Parses the file content into memory
        :param file_path: the path of the file to parse
        :return: file content as a dictionary
        """
        raise NotImplementedError()

    def has(self, entry: ConfigEntry) -> bool:
        value = self.root
        if value is None:
            return False

        for key in entry.key_path:
            value = value.get(key)
            if value is None:
                return False

        return True

    def get(self, entry: ConfigEntry) -> any:
        value = self.root
        if value is None:
            return None

        for key in entry.key_path:
            value = value.get(key)
            if value is None:
                return entry.value

        return value

    def write_reference(self, reference: dict):
        """
        Writes a reference config file
        """
        folder = self.paths[0]
        file_name = "{}_reference.{}".format(self.file_names[0], self.file_extensions[0])
        file_path = os.path.join(self.paths[0], file_name)

        os.makedirs(folder, exist_ok=True)

        self._write_reference(reference, file_path)

    def _write_reference(self, reference: dict, file_path: str):
        """
        Writes a reference config file
        :param reference: the data to write
        :param file_path: the file path to write to
        """
        pass

    def _find_config_file(self) -> str or None:
        """
        Tries to find a usable config file
        :return: file path or None
        """
        import os

        for path in self.paths:
            path = os.path.expanduser(path)
            for extension in self.file_extensions:
                for file_name in self.file_names:
                    file_path = os.path.join(path, "{}.{}".format(file_name, extension))
                    if os.path.isfile(file_path):
                        return file_path

        return None



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
from container_app_conf import ConfigBase
from container_app_conf.entry.bool import BoolConfigEntry
from container_app_conf.entry.string import StringConfigEntry
from tests import TestBase


class TestConfigBaseClashing(ConfigBase):
    clashing_key_path = ["test", "bool"]

    BOOL = BoolConfigEntry(
        key_path=clashing_key_path,
        default=False)

    STRING = StringConfigEntry(
        key_path=clashing_key_path,
        default="test"
    )


class TestClashingKeys(TestBase):

    def test_invalid_key_path_characters(self):
        def create_entry():
            BoolConfigEntry(key_path=[
                "test:"
            ])

        self.assertRaises(ValueError, create_entry)

    def test_clashing_keys(self):
        with self.assertRaises(ValueError):
            TestConfigBaseClashing()



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
import os
import pathlib

from container_app_conf import ConfigEntry


class FileConfigEntry(ConfigEntry):
    _example = "/tmp/temporary"

    def __init__(self, key_path: [str], example: any = None, description: str or None = None, default: any = None,
                 none_allowed: bool = None, check_existence: bool = False):
        """
        See ConfigEntry
        :param check_existence: whether to check for file existence
        """
        super().__init__(key_path, example, description, default, none_allowed)
        self.check_existence = check_existence

    def _value_to_type(self, value: any) -> str or None:
        """
        Tries to permissively convert the given value to a file path.
        :param value: the value to parse
        :return: the parsed file value
        """
        str_value = str(value)
        if isinstance(value, pathlib.Path):
            file = value
        else:
            file = pathlib.Path(str_value)

        if str_value.endswith(os.sep):
            raise AssertionError("File path should not end with '{}' delimiter: {}".format(os.sep, str_value))

        if file.is_dir():
            raise IsADirectoryError("Path is not a file: {}".format(str_value))

        if self.check_existence and not file.exists():
            raise FileNotFoundError("File does not exist: {}".format(value))

        return file


class DirectoryConfigEntry(ConfigEntry):
    _example = "/tmp"

    def __init__(self, key_path: [str], example: any = None, description: str or None = None, default: any = None,
                 none_allowed: bool = None, check_existence: bool = False):
        """
        See ConfigEntry
        :param check_existence: whether to check for folder existence
        """
        super().__init__(key_path, example, description, default, none_allowed)
        self.check_existence = check_existence

    def _value_to_type(self, value: any) -> str or None:
        """
        Tries to permissively convert the given value to a folder path.
        :param value: the value to parse
        :return: the parsed folder value
        """
        str_value = str(value)
        if isinstance(value, pathlib.Path):
            str_value += "/"
            directory = value
        else:
            directory = pathlib.Path(str_value)

        if not str_value.endswith(os.sep):
            raise AssertionError("Directory path should end with '{}' delimiter: {}".format(os.sep, str_value))

        if directory.is_file():
            raise NotADirectoryError("Path is not a directory: {}".format(str_value))

        if self.check_existence and not directory.exists():
            raise FileNotFoundError("directory does not exist: {}".format(value))

        return directory



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
from container_app_conf import DataSource, ConfigEntry
from container_app_conf.entry.int import IntConfigEntry
from container_app_conf.entry.string import StringConfigEntry
from container_app_conf.source.json_source import JsonSource
from container_app_conf.source.toml_source import TomlSource
from container_app_conf.source.yaml_source import YamlSource
from tests import TestBase
from tests.singleton_test import TestConfigBase2


def _key(entry: ConfigEntry) -> str:
    return "_".join(entry.key_path)


class MemoryDataSource(DataSource):
    data = {
        _key(TestConfigBase2.BOOL): True
    }

    def has(self, entry: ConfigEntry) -> bool:
        key = _key(entry)
        return key in self.data.keys()

    def get(self, entry: ConfigEntry) -> any:
        key = _key(entry)
        return self.data[key]


class MemoryDataSource2(MemoryDataSource):
    data = {
        _key(TestConfigBase2.BOOL): False
    }


class TestDataSource(TestBase):

    def test_priority(self):
        conf = TestConfigBase2(data_sources=[
            MemoryDataSource(),
            MemoryDataSource2()
        ], singleton=False)

        self.assertTrue(conf.BOOL.value)

        conf2 = TestConfigBase2(data_sources=[
            MemoryDataSource2(),
            MemoryDataSource()
        ], singleton=False)

        self.assertFalse(conf2.BOOL.value)

    def test_toml(self):
        str_entry = StringConfigEntry(
            key_path=["testing", "key1"],
            default="value"
        )
        int_entry = IntConfigEntry(
            key_path=["testing", "key2"],
            default=2
        )
        data = {
            "testing": {
                "key1": "value",
                "key2": 2,
            }
        }

        source = TomlSource("test")
        source._write_reference(data, "./test.toml")

        source.load()
        self.assertTrue(source.has(str_entry))
        self.assertEqual(source.get(str_entry), "value")
        self.assertTrue(source.has(int_entry))
        self.assertEqual(source.get(int_entry), 2)

    def test_yaml(self):
        str_entry = StringConfigEntry(
            key_path=["testing", "key1"],
            default="value"
        )
        int_entry = IntConfigEntry(
            key_path=["testing", "key2"],
            default=2
        )
        data = {
            "testing": {
                "key1": "value",
                "key2": 2,
            }
        }

        source = YamlSource("test")
        source._write_reference(data, "./test.yaml")

        source.load()
        self.assertTrue(source.has(str_entry))
        self.assertEqual(source.get(str_entry), "value")
        self.assertTrue(source.has(int_entry))
        self.assertEqual(source.get(int_entry), 2)

    def test_json(self):
        str_entry = StringConfigEntry(
            key_path=["testing", "key1"],
            default="value"
        )
        int_entry = IntConfigEntry(
            key_path=["testing", "key2"],
            default=2
        )
        data = {
            "testing": {
                "key1": "value",
                "key2": 2,
            }
        }

        source = JsonSource("test")
        source._write_reference(data, "./test.json")

        source.load()
        self.assertTrue(source.has(str_entry))
        self.assertEqual(source.get(str_entry), "value")
        self.assertTrue(source.has(int_entry))
        self.assertEqual(source.get(int_entry), 2)



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
from typing import Type

from container_app_conf import ConfigEntry


class ListConfigEntry(ConfigEntry):
    """
    Config entry allowing to specify a list of items of specified type
    """

    def __init__(self, item_type: Type[ConfigEntry], key_path: [str], example: any = None,
                 description: str or None = None,
                 default: any = None, none_allowed: bool = None,
                 item_args: dict = None,
                 delimiter: str = None):
        """
        :param item_type: the type of the ConfigEntry to use for individual list items
        :param item_args: additional constructor arguments for the item_entry
        :param delimiter: delimiter to use for splitting items when specifying the list as a string
        :param key_path: list of yaml tree entries
        :param example: example str value
        :param description: a description of this list entry
        :param default: default value
        :param none_allowed: if None is allowed for this config entry
        """
        if item_args is None:
            item_args = {}
        self._item_entry = item_type(key_path=["dummy"], **item_args)
        self.delimiter = delimiter if delimiter is not None else ","

        super().__init__(
            key_path=key_path,
            description=description,
            example=example,
            default=default,
            none_allowed=none_allowed
        )

    @property
    def example(self) -> any:
        if self.default is not None:
            return self.default

        if self._example is not None:
            return self._example

        single_example = self._item_entry.example
        return [single_example, single_example, single_example]

    def _value_to_type(self, value: any) -> [any] or None:
        """
        Tries to permissively convert the given value to a list of values.
        :param value: the value to parse
        :return: the parsed list
        """
        if not isinstance(value, list):
            value = str(value).split(self.delimiter)

        return list(map(lambda x: self._item_entry._value_to_type(x), filter(lambda x: x, value)))

    def _type_to_value(self, type: list or str) -> any:
        if isinstance(type, str):
            return type
        str_items = list(map(lambda x: self._item_entry._type_to_value(x), type))
        return str_items
